@@ -1,10 +1,10 @@
 from daft.dataframe import DataFrame
 from daft.expressions import col
 from daft.logging import setup_logger
-from daft.udf import udf
+from daft.udf import polars_udf, udf
 
-__all__ = ["DataFrame", "col", "udf"]
+__all__ = ["DataFrame", "col", "udf", "polars_udf"]
 
 setup_logger()
 
-__version__ = "0.0.8"
+__version__ = "0.0.12"
