@@ -4,20 +4,32 @@ import io
 import uuid
 from dataclasses import dataclass
 from functools import partial
-from typing import IO, Any, Callable, Dict, List, Optional, Tuple, TypeVar, Union
+from typing import (
+    IO,
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    Optional,
+    Tuple,
+    TypeVar,
+    Union,
+)
 
 import pandas
 import pyarrow as pa
 import pyarrow.parquet as papq
-from loguru import logger
 from pyarrow import csv, json
 
+from daft.context import get_context
 from daft.dataframe.schema import DataFrameSchema
 from daft.datasources import (
     CSVSourceInfo,
     InMemorySourceInfo,
     JSONSourceInfo,
     ParquetSourceInfo,
+    StorageType,
 )
 from daft.execution.operators import ExpressionType
 from daft.expressions import ColumnExpression, Expression, col
@@ -25,9 +37,6 @@ from daft.filesystem import get_filesystem_from_path
 from daft.logical import logical_plan
 from daft.logical.schema import ExpressionList
 from daft.runners.partitioning import PartitionSet
-from daft.runners.pyrunner import PyRunner
-from daft.runners.ray_runner import RayRunner
-from daft.runners.runner import Runner
 from daft.viz import DataFrameDisplay
 
 UDFReturnType = TypeVar("UDFReturnType", covariant=True)
@@ -35,11 +44,6 @@ UDFReturnType = TypeVar("UDFReturnType", covariant=True)
 ColumnInputType = Union[Expression, str]
 
 
-from daft.config import DaftSettings
-
-_RUNNER: Optional[Runner] = None
-
-
 def _sample_with_pyarrow(
     loader_func: Callable[[IO], pa.Table],
     filepath: str,
@@ -67,11 +71,21 @@ def _get_filepaths(path: str):
         return fs.ls(path)
     elif fs.isfile(path):
         return [path]
-    return fs.expand_path(path)
+    return fs.expand_path(path, recursive=True)
 
 
 class DataFrame:
+    """A Daft DataFrame is a table of data. It has columns, where each column has a type and the same
+    number of items (rows) as all other columns.
+    """
+
     def __init__(self, plan: logical_plan.LogicalPlan) -> None:
+        """Constructs a DataFrame according to a given LogicalPlan. Users are expected instead to call
+        the classmethods on DataFrame to create a DataFrame.
+
+        Args:
+            plan: LogicalPlan describing the steps required to arrive at this DataFrame
+        """
         self._plan = plan
         self._result: Optional[PartitionSet] = None
 
@@ -83,11 +97,19 @@ class DataFrame:
         """
         return self._plan
 
+    def num_partitions(self) -> int:
+        return self._plan.num_partitions()
+
     def schema(self) -> DataFrameSchema:
+        """Returns the DataFrameSchema of the DataFrame, which provides information about each column
+
+        Returns:
+            DataFrameSchema: schema of the DataFrame
+        """
         return DataFrameSchema.from_expression_list(self._plan.schema())
 
     def column_names(self) -> List[str]:
-        """returns column names of DataFrame as a list of strings.
+        """Returns column names of DataFrame as a list of strings.
 
         Returns:
             List[str]: Column names of this DataFrame.
@@ -95,6 +117,17 @@ class DataFrame:
         return [expr.name() for expr in self._plan.schema()]
 
     def show(self, n: int = -1) -> DataFrameDisplay:
+        """Executes and displays the executed dataframe as a table
+
+        Args:
+            n: number of rows to show. Defaults to -1.
+
+        Returns:
+            DataFrameDisplay: object that has a rich tabular display
+
+            .. NOTE::
+                This call is **blocking** and will execute the DataFrame when called
+        """
         df = self
         if n != -1:
             df = df.limit(n)
@@ -113,6 +146,17 @@ class DataFrame:
 
     @classmethod
     def from_pylist(cls, data: List[Dict[str, Any]]) -> DataFrame:
+        """Creates a DataFrame from a list of dictionaries
+
+        Example:
+            >>> df = DataFrame.from_pylist([{"foo": 1}, {"foo": 2}])
+
+        Args:
+            data: list of dictionaries, where each key is a column name
+
+        Returns:
+            DataFrame: DataFrame created from list of dictionaries
+        """
         if not data:
             raise ValueError("Unable to create DataFrame from empty list")
         schema = ExpressionList(
@@ -133,11 +177,14 @@ class DataFrame:
     def from_pydict(cls, data: Dict[str, Any]) -> DataFrame:
         """Creates a DataFrame from an In-Memory Data columnar source that is passed in as `data`.
 
+        Example:
+            >>> df = DataFrame.from_pydict({"foo": [1, 2]})
+
         Args:
-            data (Dict[str, Any]): Key -> Sequence[item] of data. Each Key is created as a column.
+            data: Key -> Sequence[item] of data. Each Key is created as a column.
 
         Returns:
-            DataFrame: parsed DataFrame that will read from an In-Memory Data Source when collected.
+            DataFrame: DataFrame created from dictionary of columns
         """
         schema = ExpressionList(
             [ColumnExpression(header, expr_type=ExpressionType.from_py_type(type(data[header][0]))) for header in data]
@@ -157,6 +204,12 @@ class DataFrame:
     ) -> DataFrame:
         """Creates a DataFrame from line-delimited JSON file(s)
 
+        Example:
+            >>> df = DataFrame.from_json("/path/to/file.json")
+            >>> df = DataFrame.from_json("/path/to/directory")
+            >>> df = DataFrame.from_json("/path/to/files-*.json")
+            >>> df = DataFrame.from_json("s3://path/to/files-*.json")
+
         Args:
             path (str): Path to JSON files (allows for wildcards)
 
@@ -188,6 +241,12 @@ class DataFrame:
     ) -> DataFrame:
         """Creates a DataFrame from CSV file(s)
 
+        Example:
+            >>> df = DataFrame.from_csv("/path/to/file.csv")
+            >>> df = DataFrame.from_csv("/path/to/directory")
+            >>> df = DataFrame.from_csv("/path/to/files-*.csv")
+            >>> df = DataFrame.from_csv("s3://path/to/files-*.csv")
+
         Args:
             path (str): Path to CSV (allows for wildcards)
             has_headers (bool): Whether the CSV has a header or not, defaults to True
@@ -234,6 +293,12 @@ class DataFrame:
     def from_parquet(cls, path: str) -> DataFrame:
         """Creates a DataFrame from Parquet file(s)
 
+        Example:
+            >>> df = DataFrame.from_parquet("/path/to/file.parquet")
+            >>> df = DataFrame.from_parquet("/path/to/directory")
+            >>> df = DataFrame.from_parquet("/path/to/files-*.parquet")
+            >>> df = DataFrame.from_parquet("s3://path/to/files-*.parquet")
+
         Args:
             path (str): Path to Parquet file (allows for wildcards)
 
@@ -266,21 +331,126 @@ class DataFrame:
         )
         return cls(plan)
 
+    @classmethod
+    def from_files(cls, path: str) -> DataFrame:
+        """Creates a DataFrame from files in storage, where each file is one row of the DataFrame
+
+        Args:
+            path (str): path to files on disk (allows wildcards)
+
+        Returns:
+            DataFrame: DataFrame containing the path to each file as a row, along with other metadata
+                parsed from the provided filesystem
+        """
+        fs = get_filesystem_from_path(path)
+        file_details = fs.glob(path, detail=True)
+        return cls.from_pylist(list(file_details.values()))
+
+    ###
+    # Write methods
+    ###
+
+    def write_parquet(
+        self, root_dir: str, compression: str = "snappy", partition_cols: Optional[List[ColumnInputType]] = None
+    ) -> DataFrame:
+        """Writes the DataFrame to parquet files using a `root_dir` and randomly generated UUIDs as the filepath and returns the filepaths.
+        Currently generates a parquet file per partition unless `partition_cols` are used, then the number of files can equal the number of partitions times the number of values of partition col.
+
+        Args:
+            root_dir (str): root file path to write parquet files to.
+            compression (str, optional): compression algorithm. Defaults to "snappy".
+            partition_cols (Optional[List[ColumnInputType]], optional): How to subpartition each partition further. Currently only supports ColumnExpressions with any calls. Defaults to None.
+
+        Returns:
+            DataFrame: The filenames that were written out as strings.
+
+            .. NOTE::
+                This call is **blocking** and will execute the DataFrame when called
+        """
+        cols: Optional[ExpressionList] = None
+        if partition_cols is not None:
+            cols = self.__column_input_to_expression(tuple(partition_cols))
+            for c in cols:
+                assert isinstance(c, ColumnExpression), "we cant support non ColumnExpressions for partition writing"
+            df = self.repartition(self.num_partitions(), *cols)
+        else:
+            df = self
+        plan = logical_plan.FileWrite(
+            df._plan,
+            root_dir=root_dir,
+            partition_cols=cols,
+            storage_type=StorageType.PARQUET,
+            compression=compression,
+        )
+
+        # Block and write, then retrieve data and return a new disconnected DataFrame
+        write_df = DataFrame(plan)
+        write_df.collect()
+        assert write_df._result is not None
+        data = write_df._result.to_pydict()
+        return DataFrame.from_pydict(data)
+
+    def write_csv(self, root_dir: str, partition_cols: Optional[List[ColumnInputType]] = None) -> DataFrame:
+        """Writes the DataFrame to CSV files using a `root_dir` and randomly generated UUIDs as the filepath and returns the filepaths.
+        Currently generates a csv file per partition unless `partition_cols` are used, then the number of files can equal the number of partitions times the number of values of partition col.
+
+        Args:
+            root_dir (str): root file path to write parquet files to.
+            compression (str, optional): compression algorithm. Defaults to "snappy".
+            partition_cols (Optional[List[ColumnInputType]], optional): How to subpartition each partition further. Currently only supports ColumnExpressions with any calls. Defaults to None.
+
+        Returns:
+            DataFrame: The filenames that were written out as strings.
+
+            .. NOTE::
+                This call is **blocking** and will execute the DataFrame when called
+        """
+        cols: Optional[ExpressionList] = None
+        if partition_cols is not None:
+            cols = self.__column_input_to_expression(tuple(partition_cols))
+            for c in cols:
+                assert isinstance(c, ColumnExpression), "we cant support non ColumnExpressions for partition writing"
+            df = self.repartition(self.num_partitions(), *cols)
+        else:
+            df = self
+        plan = logical_plan.FileWrite(
+            df._plan,
+            root_dir=root_dir,
+            partition_cols=cols,
+            storage_type=StorageType.CSV,
+        )
+
+        # Block and write, then retrieve data and return a new disconnected DataFrame
+        write_df = DataFrame(plan)
+        write_df.collect()
+        assert write_df._result is not None
+        data = write_df._result.to_pydict()
+        return DataFrame.from_pydict(data)
+
     ###
     # DataFrame operations
     ###
 
-    def __column_input_to_expression(self, columns: Tuple[ColumnInputType, ...]) -> ExpressionList:
+    def __column_input_to_expression(self, columns: Iterable[ColumnInputType]) -> ExpressionList:
         expressions = [col(c) if isinstance(c, str) else c for c in columns]
         return ExpressionList(expressions)
 
     def select(self, *columns: ColumnInputType) -> DataFrame:
         """Creates a new DataFrame that `selects` that columns that are passed in from the current DataFrame.
-        columns can be:
-        * names of columns as strings. `df.select('x', 'y')`
-        * names of columns as expressions. `df.select(col('x'), col('y'))`
-        * call expressions. `df.select(col('x') * col('y'))`
-        * any of the above mixed in a call. `df.select('x', col('y'), col('z') + 1)`
+
+        Example:
+
+            >>> # names of columns as strings
+            >>> df = df.select('x', 'y')
+            >>>
+            >>> # names of columns as expressions
+            >>> df = df.select(col('x'), col('y'))
+            >>>
+            >>> # call expressions
+            >>> df = df.select(col('x') * col('y'))
+            >>>
+            >>> # any mix of the above
+            >>> df = df.select('x', col('y'), col('z') + 1)
 
         Args:
             *columns (Union[str, Expression]): columns to select from the current DataFrame
@@ -294,7 +464,9 @@ class DataFrame:
 
     def distinct(self) -> DataFrame:
         """Computes unique rows, dropping duplicates.
-        `unique_df = df.distinct()`
+
+        Example:
+            >>> unique_df = df.distinct()
 
         Returns:
             DataFrame: DataFrame that has only  unique rows.
@@ -308,7 +480,9 @@ class DataFrame:
     def exclude(self, *names: str) -> DataFrame:
         """Drops columns from the current DataFrame by name.
         This is equivalent of performing a select with all the columns but the ones excluded.
-        `df_without_x = df.exclude('x')`
+
+        Example:
+            >>> df_without_x = df.exclude('x')
 
         Args:
             *names (str): names to exclude
@@ -323,7 +497,9 @@ class DataFrame:
     def where(self, predicate: Expression) -> DataFrame:
         """Filters rows via a predicate expression.
         similar to SQL style `where`.
-        `filtered_df = df.where((col('x') < 10) & (col('y') == 10))`
+
+        Example:
+            >>> filtered_df = df.where((col('x') < 10) & (col('y') == 10))
 
         Args:
             predicate (Expression): expression that keeps row if evaluates to True.
@@ -337,7 +513,9 @@ class DataFrame:
     def with_column(self, column_name: str, expr: Expression) -> DataFrame:
         """Adds a column to the current DataFrame with an Expression.
         This is equivalent to performing a `select` with all the current columns and the new one.
-        `new_df = df.with_column('x+1', col('x') + 1)`
+
+        Example:
+            >>> new_df = df.with_column('x+1', col('x') + 1)
 
         Args:
             column_name (str): name of new column
@@ -348,31 +526,43 @@ class DataFrame:
         """
         prev_schema_as_cols = self._plan.schema().to_column_expressions()
         projection = logical_plan.Projection(
-            self._plan, prev_schema_as_cols.union(ExpressionList([expr.alias(column_name)]))
+            self._plan, prev_schema_as_cols.union(ExpressionList([expr.alias(column_name)]), other_override=True)
         )
         return DataFrame(projection)
 
-    def sort(self, column: ColumnInputType, desc: bool = False) -> DataFrame:
+    def sort(
+        self, by: Union[ColumnInputType, List[ColumnInputType]], desc: Union[bool, List[bool]] = False
+    ) -> DataFrame:
         """Sorts DataFrame globally according to column.
-        `sorted_df = df.sort(col('x') + col('y'))`
+
+        Example:
+            >>> sorted_df = df.sort(col('x') + col('y'))
+            >>> sorted_df = df.sort([col('x'), col('y')], desc=[False, True])
+            >>> sorted_df = df.sort(['z', col('x'), col('y')], desc=[True, False, True])
 
         Note:
             * Since this a global sort, this requires an expensive repartition which can be quite slow.
-            * Only single expression sorts are currently implemented.
+            * Supports multicolumn sorts and can have unique `descending` flag per column.
         Args:
-            column (ColumnInputType): column to sort by. Can be `str` or expression.
-            desc (bool, optional): Sort by descending order. Defaults to False.
+            column (Union[ColumnInputType, List[ColumnInputType]]): column to sort by. Can be `str` or expression as well as a list of either.
+            desc (Union[bool, List[bool]), optional): Sort by descending order. Defaults to False.
 
         Returns:
             DataFrame: Sorted DataFrame.
         """
-        sort = logical_plan.Sort(self._plan, self.__column_input_to_expression((column,)), desc=desc)
+        if not isinstance(by, list):
+            by = [
+                by,
+            ]
+        sort = logical_plan.Sort(self._plan, self.__column_input_to_expression(by), descending=desc)
         return DataFrame(sort)
 
     def limit(self, num: int) -> DataFrame:
         """Limits the rows returned by the DataFrame via a `head` operation.
         This is similar to how `limit` works in SQL.
-        `df_limited = df.limit(10) # returns 10 rows`
+
+        Example:
+            >>> df_limited = df.limit(10) # returns 10 rows
 
         Args:
             num (int): maximum rows to allow.
@@ -385,12 +575,14 @@ class DataFrame:
         return DataFrame(global_limit)
 
     def repartition(self, num: int, *partition_by: ColumnInputType) -> DataFrame:
-        """repartitions DataFrame to `num` partitions.
-        if columns are passed in, then DataFrame will be repartitioned by those,
-            otherwise random repartitioning will occur.
-        `random_repart_df = df.repartition(4)`
+        """Repartitions DataFrame to `num` partitions.
 
-        `part_by_df = df.repartition(4, 'x', col('y') + 1)`
+        If columns are passed in, then DataFrame will be repartitioned by those, otherwise
+        random repartitioning will occur.
+
+        Example:
+            >>> random_repart_df = df.repartition(4)
+            >>> part_by_df = df.repartition(4, 'x', col('y') + 1)
 
         Args:
             num (int): number of target partitions.
@@ -421,9 +613,9 @@ class DataFrame:
         """Joins left (self) DataFrame on the right on a set of keys.
         Key names can be the same or different for left and right DataFrame.
 
-        Note: Although self joins are supported,
-            we currently duplicate the logical plan for the right side and recompute the entire tree.
-            Caching for this is on the roadmap.
+        .. NOTE::
+            Although self joins are supported, we currently duplicate the logical plan for the right side
+            and recompute the entire tree. Caching for this is on the roadmap.
 
         Args:
             other (DataFrame): the right DataFrame to join on.
@@ -456,6 +648,53 @@ class DataFrame:
         )
         return DataFrame(join_op)
 
+    def explode(self, *columns: ColumnInputType) -> DataFrame:
+        """Explodes a List column, where every element in each row's List becomes its own row, and all
+        other columns in the DataFrame are duplicated across rows.
+
+        If multiple columns are specified, each row must contain the same number of
+        items in each specified column.
+
+        Exploding Null values or empty lists will create a single Null entry (see example below).
+
+        Example:
+            >>> df = DataFrame.from_pydict({
+            >>>     "x": [[1], [2, 3]],
+            >>>     "y": [["a"], ["b", "c"]],
+            >>>     "z": [1.0, 2.0],
+            >>> ]})
+            >>>
+            >>> df.explode(col("x"), col("y"))
+            >>>
+            >>> # +------+-----------+-----+      +------+------+-----+
+            >>> # | x    | y         | z   |      |  x   |  y   | z   |
+            >>> # +------+-----------+-----+      +------+------+-----+
+            >>> # |[1]   | ["a"]     | 1.0 |      |  1   | "a"  | 1.0 |
+            >>> # +------+-----------+-----+  ->  +------+------+-----+
+            >>> # |[2, 3]| ["b", "c"]| 2.0 |      |  2   | "b"  | 2.0 |
+            >>> # +------+-----------+-----+      +------+------+-----+
+            >>> # |[]    | []        | 3.0 |      |  3   | "c"  | 2.0 |
+            >>> # +------+-----------+-----+      +------+------+-----+
+            >>> # |None  | None      | 4.0 |      | None | None | 3.0 |
+            >>> # +------+-----------+-----+      +------+------+-----+
+            >>> #                                 | None | None | 4.0 |
+            >>> #                                 +------+------+-----+
+
+        Args:
+            *columns (ColumnInputType): columns to explode
+
+        Returns:
+            DataFrame: DataFrame with exploded column
+        """
+        if len(columns) < 1:
+            raise ValueError("At least one column to explode must be specified")
+        exprs_to_explode = self.__column_input_to_expression(columns)
+        explode_op = logical_plan.Explode(
+            self._plan,
+            exprs_to_explode,
+        )
+        return DataFrame(explode_op)
+
     def _agg(self, to_agg: List[Tuple[ColumnInputType, str]], group_by: Optional[ExpressionList] = None) -> DataFrame:
         assert len(to_agg) > 0, "no columns to aggregate."
         exprs_to_agg = self.__column_input_to_expression(tuple(e for e, _ in to_agg))
@@ -596,8 +835,9 @@ class DataFrame:
         Returns:
             DataFrame: DataFrame with cached results.
         """
+        context = get_context()
         if self._result is None:
-            self._result = self._get_runner().run(self._plan)
+            self._result = context.runner().run(self._plan)
         return self
 
     def to_pandas(self) -> pandas.DataFrame:
@@ -606,6 +846,9 @@ class DataFrame:
 
         Returns:
             pandas.DataFrame: pandas DataFrame converted from a Daft DataFrame
+
+            .. NOTE::
+                This call is **blocking** and will execute the DataFrame when called
         """
         self.collect()
         assert self._result is not None
@@ -614,18 +857,6 @@ class DataFrame:
         self._result = None
         return pd_df
 
-    def _get_runner(self) -> Runner:
-        global _RUNNER
-        if _RUNNER is not None:
-            return _RUNNER
-        if DaftSettings.DAFT_RUNNER.upper() == "RAY":
-            logger.info("Using RayRunner")
-            _RUNNER = RayRunner()
-        else:
-            logger.info("Using PyRunner")
-            _RUNNER = PyRunner()
-        return _RUNNER
-
 
 @dataclass
 class GroupedDataFrame:
@@ -633,7 +864,7 @@ class GroupedDataFrame:
     group_by: ExpressionList
 
     def sum(self, *cols: ColumnInputType) -> DataFrame:
-        """performs grouped sum on this Grouped DataFrame.
+        """Perform grouped sum on this GroupedDataFrame.
 
         Args:
             *cols (Union[str, Expression]): columns to sum
@@ -644,7 +875,7 @@ class GroupedDataFrame:
         return self.df._agg([(c, "sum") for c in cols], group_by=self.group_by)
 
     def mean(self, *cols: ColumnInputType) -> DataFrame:
-        """performs grouped mean on this Grouped DataFrame.
+        """Performs grouped mean on this GroupedDataFrame.
 
         Args:
             *cols (Union[str, Expression]): columns to mean
@@ -656,9 +887,15 @@ class GroupedDataFrame:
         return self.df._agg([(c, "mean") for c in cols], group_by=self.group_by)
 
     def agg(self, to_agg: List[Tuple[ColumnInputType, str]]) -> DataFrame:
-        """performed aggregations on this grouped DataFrame.
-        Allows for mixed aggregations.
-        `df = df.groupby('x').agg([('x', 'sum'), ('x', 'mean'), ('y', 'min'), ('y', 'max')])`
+        """Perform aggregations on this GroupedDataFrame. Allows for mixed aggregations.
+
+        Example:
+        >>> df = df.groupby('x').agg([
+        >>>     ('x', 'sum'),
+        >>>     ('x', 'mean'),
+        >>>     ('y', 'min'),
+        >>>     ('y', 'max'),
+        >>> ])
 
         Args:
             to_agg (List[Tuple[ColumnInputType, str]]): list of (column, agg_type)
