@@ -1,6 +1,7 @@
 from __future__ import annotations
 
 import itertools
+import warnings
 from abc import abstractmethod
 from copy import deepcopy
 from functools import partial, partialmethod
@@ -8,24 +9,24 @@ from typing import (
     Any,
     Callable,
     Dict,
-    Generic,
     List,
     NewType,
     Optional,
+    Sequence,
     Tuple,
     Type,
     TypeVar,
     cast,
 )
 
+import numpy as np
+import pandas as pd
+import polars as pl
+import pyarrow as pa
+
+from daft.errors import ExpressionTypeError
 from daft.execution import url_operators
-from daft.execution.operators import (
-    CompositeExpressionType,
-    ExpressionOperator,
-    ExpressionType,
-    OperatorEnum,
-    PythonExpressionType,
-)
+from daft.execution.operators import ExpressionOperator, OperatorEnum
 from daft.internal.treenode import TreeNode
 from daft.resource_request import ResourceRequest
 from daft.runners.blocks import (
@@ -34,19 +35,41 @@ from daft.runners.blocks import (
     PyListDataBlock,
     zip_blocks_as_py,
 )
+from daft.types import ExpressionType, PrimitiveExpressionType
 
 
 def col(name: str) -> ColumnExpression:
+    """Selects a column with a given name
+
+    Example:
+        >>> col("x")
+
+    Args:
+        name: Name of column
+
+    Returns:
+        ColumnExpression: Expression representing the selected column
+    """
     return ColumnExpression(name)
 
 
 def lit(val: Any) -> LiteralExpression:
+    """Selects a column with every item being the provided value
+
+    Example:
+        >>> col("x") + lit(1)
+
+    Args:
+        val: value of column
+
+    Returns:
+        LiteralExpression: Expression representing the value provided
+    """
     return LiteralExpression(val)
 
 
 _COUNTER = itertools.count()
 ColID = NewType("ColID", int)
-DataBlockValueType = TypeVar("DataBlockValueType", bound=DataBlock)
 
 
 class ExpressionExecutor:
@@ -69,7 +92,7 @@ class ExpressionExecutor:
             # Use a PyListDataBlock evaluator if any of the args are Python types
             op_evaluator = (
                 PyListDataBlock.evaluator
-                if any([isinstance(arg.resolved_type(), PythonExpressionType) for arg in expr._args])
+                if any(isinstance(block, PyListDataBlock) for block in eval_args)
                 else ArrowDataBlock.evaluator
             )
             op = expr._operator
@@ -217,22 +240,17 @@ class Expression(TreeNode["Expression"]):
     __pos__ = partialmethod(_unary_op, OperatorEnum.POSITIVE)
     __abs__ = partialmethod(_unary_op, OperatorEnum.ABS)
 
+    # Symbolic - not used during evaluation but useful for type resolution on the Expression
     _sum = partialmethod(_unary_op, OperatorEnum.SUM)
     _count = partialmethod(_unary_op, OperatorEnum.COUNT)
     _mean = partialmethod(_unary_op, OperatorEnum.MEAN)
     _min = partialmethod(_unary_op, OperatorEnum.MIN)
     _max = partialmethod(_unary_op, OperatorEnum.MAX)
+    _explode = partialmethod(_unary_op, OperatorEnum.EXPLODE)
+
     # Logical
     __invert__ = partialmethod(_unary_op, OperatorEnum.INVERT)
 
-    # # function
-    # def map(self, func: Callable) -> Expression:
-    #     return self._unary_op(
-    #         # TODO(jay): Figure out how to do this operator correctly, we use a placeholder here for now
-    #         None,  # type: ignore
-    #         func,
-    #     )
-
     # BinaryOps
 
     # Arithmetic
@@ -272,19 +290,67 @@ class Expression(TreeNode["Expression"]):
         raise NotImplementedError()
 
     def alias(self, name: str) -> Expression:
+        """Gives the expression a new name, which is its column's name in the DataFrame schema and the name
+        by which subsequent expressions can refer to the results of this expression.
+
+        Example:
+            >>> col("x").alias("y")
+
+        Args:
+            name: New name for expression
+
+        Returns:
+            Expression: Renamed expression
+        """
         return AliasExpression(self, name)
 
-    def as_py(self, type_: Type) -> Expression:
+    def as_py(self, type_: Type) -> AsPyExpression:
+        """Treats every value on the given expression as an object of the specified type. Users can then call
+        methods on this object, which will be translated into a method call on every value on the Expression.
+
+        Example:
+            >>> # call .resize(28, 28) on every item in the expression
+            >>> col("images").as_py(Image).resize(28, 28)
+            >>>
+            >>> # get the 0th element of every item in the expression
+            >>> col("tuples").as_py(tuple)[0]
+
+        Args:
+            type_ (Type): type of each item in the expression
+
+        Returns:
+            AsPyExpression: A special Expression that records any method calls that a user runs on it, applying the method call to each item in the expression.
+        """
         return AsPyExpression(self, ExpressionType.from_py_type(type_))
 
-    def apply(self, func: Callable, return_type: Optional[Type] = None):
+    def apply(self, func: Callable, return_type: Optional[Type] = None) -> Expression:
+        """Apply a function on a given expression
+
+        Example:
+            >>> def f(x_val: str) -> int:
+            >>>     return int(x_val) if x_val.isnumeric() else 0
+            >>>
+            >>> col("x").apply(f, return_type=int)
+
+        Args:
+            func (Callable): Function to run per value of the expression
+            return_type (Optional[Type], optional): Return type of the function that was ran. Defaults to None.
+
+        Returns:
+            Expression: New expression after having run the function on the expression
+        """
+        if return_type is None:
+            warnings.warn(
+                "No `return_type` keyword argument specified in `.apply`. It is highly recommended to "
+                "specify a return_type for Daft to perform optimizations and for access to vectorized expression operators."
+            )
+
         expression_type = (
             ExpressionType.from_py_type(return_type) if return_type is not None else ExpressionType.python_object()
         )
 
         def apply_func(f, data):
-            results = list(map(f, data.iter_py()))
-            return DataBlock.make_block(results)
+            return list(map(f, data.iter_py()))
 
         return UdfExpression(
             partial(apply_func, func),
@@ -326,19 +392,54 @@ class Expression(TreeNode["Expression"]):
 
         return True
 
-    def if_else(self, other1: Expression, other2: Expression) -> Expression:
+    def if_else(self, if_true: Expression, if_false: Expression) -> Expression:
+        """Choose values from `if_true` and `if_false` based on the current expression as the condition.
+
+        Example:
+            >>> # x = [2, 2, 2]
+            >>> # y = [1, 2, 3]
+            >>> # a = ["a", "a", "a"]
+            >>> # b = ["b", "b", "b"]
+            >>> # if_else_result = ["a", "b", "b"]
+            >>> (col("x") > col("y")).if_else(col("a"), col("b"))
+
+        Args:
+            if_true (Expression): Values to choose if condition is true
+            if_false (Expression): Values to choose if condition is false
+
+        Returns:
+            Expression: New expression where values are chosen from `if_true` and `if_false`.
+        """
         return CallExpression(
             OperatorEnum.IF_ELSE,
-            (self, other1, other2),
+            (self, if_true, if_false),
         )
 
     def is_null(self) -> Expression:
+        """Checks if values in the Expression are Null (i.e. they are missing)
+
+        Example:
+            >>> # [1., None, NaN] -> [False, True, False]
+            >>> col("x").is_null()
+
+        Returns:
+            Expression: LOGICAL Expression indicating whether values are missing
+        """
         return CallExpression(
             OperatorEnum.IS_NULL,
             (self,),
         )
 
     def is_nan(self) -> Expression:
+        """Checks if values in the Expression are NaN (i.e. they are invalid)
+
+        Example:
+            >>> # [1., None, NaN] -> [False, False, True]
+            >>> col("x").is_nan()
+
+        Returns:
+            Expression: LOGICAL Expression indicating whether values are invalid.
+        """
         return CallExpression(
             OperatorEnum.IS_NAN,
             (self,),
@@ -379,30 +480,6 @@ class LiteralExpression(Expression):
         return isinstance(other, LiteralExpression) and self._value == other._value
 
 
-class MultipleReturnSelectExpression(Expression):
-    def __init__(self, expr: UdfExpression, n: int) -> None:
-        super().__init__()
-        self._register_child(expr)
-        self._n = n
-
-    def resolved_type(self) -> Optional[ExpressionType]:
-        call_resolved_type = self._expr.resolved_type()
-        if call_resolved_type is None:
-            return None
-        assert isinstance(call_resolved_type, CompositeExpressionType)
-        return call_resolved_type.args[self._n]  # type: ignore
-
-    @property
-    def _expr(self) -> UdfExpression:
-        return cast(UdfExpression, self._children()[0])
-
-    def _display_str(self) -> str:
-        return f"{self._expr}[{self._n}]"
-
-    def _is_eq_local(self, other: Expression) -> bool:
-        return isinstance(other, MultipleReturnSelectExpression) and self._n == other._n
-
-
 class CallExpression(Expression):
     def __init__(
         self,
@@ -428,7 +505,7 @@ class CallExpression(Expression):
                 if operator.nargs == 2
                 else f"{operator_symbol}({', '.join([str(arg) for arg in self._args])})"
             )
-            raise TypeError(f"Unable to resolve type for operation: {op_pretty_print}")
+            raise ExpressionTypeError(f"Unable to resolve type for operation: {op_pretty_print}")
 
         return ret_type
 
@@ -454,10 +531,10 @@ class CallExpression(Expression):
         )
 
 
-class UdfExpression(Expression, Generic[DataBlockValueType]):
+class UdfExpression(Expression):
     def __init__(
         self,
-        func: Callable[..., DataBlockValueType],
+        func: Callable[..., Sequence],
         func_ret_type: ExpressionType,
         func_args: Tuple,
         func_kwargs: Dict[str, Any],
@@ -494,8 +571,35 @@ class UdfExpression(Expression, Generic[DataBlockValueType]):
             return f"{func_name}({args}, {kwargs})"
         return f"{func_name}({args})"
 
-    def eval_blocks(self, *args: DataBlockValueType, **kwargs: DataBlockValueType) -> DataBlockValueType:
-        return self._func(*args, **kwargs)
+    def eval_blocks(self, *args: DataBlock, **kwargs: DataBlock) -> DataBlock:
+        results = self._func(*args, **kwargs)
+
+        if ExpressionType.is_py(self._func_ret_type):
+            return PyListDataBlock(results if isinstance(results, list) else list(results))
+
+        # Convert these user-provided types to pa.ChunkedArray for making blocks
+        if isinstance(results, pa.Array):
+            results = pa.chunked_array([results])
+        elif isinstance(results, pl.Series):
+            results = pa.chunked_array([results.to_arrow()])
+        elif isinstance(results, np.ndarray):
+            results = pa.chunked_array([pa.array(results)])
+        elif isinstance(results, pd.Series):
+            results = pa.chunked_array([pa.array(results)])
+        elif isinstance(results, list):
+            results = pa.chunked_array([pa.array(results)])
+
+        if not isinstance(results, pa.ChunkedArray):
+            raise NotImplementedError(f"Cannot make block for data of type: {type(results)}")
+
+        # Explcitly cast results of the UDF here for these primitive types:
+        #   1. Ensures that all blocks across all partitions will have the same underlying Arrow type after the UDF
+        #   2. Ensures that any null blocks from empty partitions or partitions with all nulls have the correct type
+        assert isinstance(self._func_ret_type, PrimitiveExpressionType)
+        expected_arrow_type = self._func_ret_type.to_arrow_type()
+        results = results.cast(expected_arrow_type)
+
+        return ArrowDataBlock(results)
 
     def _is_eq_local(self, other: Expression) -> bool:
         return (
@@ -620,7 +724,7 @@ class AsPyExpression(Expression):
                     **{kw: val for kw, val in zip(kwargs.keys(), *vals[len(args) + 1 :])},
                 )
                 results.append(result)
-            return DataBlock.make_block(results)
+            return results
 
         return UdfExpression(
             f,
@@ -639,7 +743,7 @@ class AsPyExpression(Expression):
                 method = getattr(python_cls, attr_name)
                 result = method(expr_val, key_val)
                 results.append(result)
-            return DataBlock.make_block(results)
+            return results
 
         return UdfExpression(
             __getitem__,
@@ -691,28 +795,71 @@ class UrlMethodAccessor(BaseMethodAccessor):
 
 class StringMethodAccessor(BaseMethodAccessor):
     def contains(self, pattern: str) -> CallExpression:
-        """Checks whether each string contains the given pattern in a string column"""
+        """Checks whether each string contains the given pattern in a string column
+
+        Example:
+            >>> col("x").str.contains("foo")
+
+        Args:
+            pattern (str): pattern to search for
+
+        Returns:
+            CallExpression: a LOGICAL expression indicating whether each value contains the provided pattern
+        """
+        if not isinstance(pattern, str):
+            raise ExpressionTypeError(f"Expected pattern to be a Python string, received: {pattern}")
         return CallExpression(
             OperatorEnum.STR_CONTAINS,
             (self._expr, pattern),
         )
 
     def endswith(self, pattern: str) -> CallExpression:
-        """Checks whether each string ends with the given pattern in a string column"""
+        """Checks whether each string ends with the given pattern in a string column
+
+        Example:
+            >>> col("x").str.endswith("foo")
+
+        Args:
+            pattern (str): pattern to search for
+
+        Returns:
+            CallExpression: a LOGICAL expression indicating whether each value ends with the provided pattern
+        """
+        if not isinstance(pattern, str):
+            raise ExpressionTypeError(f"Expected pattern to be a Python string, received: {pattern}")
         return CallExpression(
             OperatorEnum.STR_ENDSWITH,
             (self._expr, pattern),
         )
 
     def startswith(self, pattern: str) -> CallExpression:
-        """Checks whether each string starts with the given pattern in a string column"""
+        """Checks whether each string starts with the given pattern in a string column
+
+        Example:
+            >>> col("x").str.startswith("foo")
+
+        Args:
+            pattern (str): pattern to search for
+
+        Returns:
+            CallExpression: a LOGICAL expression indicating whether each value starts with the provided pattern
+        """
+        if not isinstance(pattern, str):
+            raise ExpressionTypeError(f"Expected pattern to be a Python string, received: {pattern}")
         return CallExpression(
             OperatorEnum.STR_STARTSWITH,
             (self._expr, pattern),
         )
 
     def length(self) -> CallExpression:
-        """Retrieves the length for of a UTF-8 string column"""
+        """Retrieves the length for of a UTF-8 string column
+
+        Example:
+            >>> col("x").str.length()
+
+        Returns:
+            CallExpression: an INTEGER expression with the length of each string
+        """
         return CallExpression(
             OperatorEnum.STR_LENGTH,
             (self._expr,),
@@ -721,28 +868,56 @@ class StringMethodAccessor(BaseMethodAccessor):
 
 class DatetimeMethodAccessor(BaseMethodAccessor):
     def day(self) -> CallExpression:
-        """Retrieves the day for a datetime column"""
+        """Retrieves the day for a datetime column
+
+        Example:
+            >>> col("x").dt.day()
+
+        Returns:
+            CallExpression: an INTEGER expression with just the day extracted from a datetime column
+        """
         return CallExpression(
             OperatorEnum.DT_DAY,
             (self._expr,),
         )
 
     def month(self) -> CallExpression:
-        """Retrieves the month for a datetime column"""
+        """Retrieves the month for a datetime column
+
+        Example:
+            >>> col("x").dt.month()
+
+        Returns:
+            CallExpression: an INTEGER expression with just the month extracted from a datetime column
+        """
         return CallExpression(
             OperatorEnum.DT_MONTH,
             (self._expr,),
         )
 
     def year(self) -> CallExpression:
-        """Retrieves the year for a datetime column"""
+        """Retrieves the year for a datetime column
+
+        Example:
+            >>> col("x").dt.year()
+
+        Returns:
+            CallExpression: an INTEGER expression with just the year extracted from a datetime column
+        """
         return CallExpression(
             OperatorEnum.DT_YEAR,
             (self._expr,),
         )
 
     def day_of_week(self) -> CallExpression:
-        """Retrieves the day of the week for a datetime column, starting at 0 for Monday and ending at 6 for Sunday"""
+        """Retrieves the day of the week for a datetime column, starting at 0 for Monday and ending at 6 for Sunday
+
+        Example:
+            >>> col("x").dt.day_of_week()
+
+        Returns:
+            CallExpression: an INTEGER expression with just the day_of_week extracted from a datetime column
+        """
         return CallExpression(
             OperatorEnum.DT_DAY_OF_WEEK,
             (self._expr,),
