@@ -4,14 +4,16 @@ import itertools
 from abc import abstractmethod
 from dataclasses import dataclass
 from enum import Enum, IntEnum
-from typing import Any, List, Optional, Tuple
+from typing import Any, Generic, List, Optional, Tuple, TypeVar, Union
 
-from daft.datasources import SourceInfo
-from daft.execution.operators import ExpressionType
+from daft.datasources import SourceInfo, StorageType
 from daft.expressions import ColumnExpression, Expression
 from daft.internal.treenode import TreeNode
+from daft.logical.map_partition_ops import ExplodeOp, MapPartitionOp
 from daft.logical.schema import ExpressionList
 from daft.resource_request import ResourceRequest
+from daft.runners.partitioning import vPartition
+from daft.types import ExpressionType
 
 
 class OpLevel(IntEnum):
@@ -51,6 +53,7 @@ class LogicalPlan(TreeNode["LogicalPlan"]):
             isinstance(other, LogicalPlan)
             and self._local_eq(other)
             and self.schema() == other.schema()
+            and self.partition_spec() == other.partition_spec()
             and self.num_partitions() == other.num_partitions()
             and all(
                 [self_child.is_eq(other_child) for self_child, other_child in zip(self._children(), other._children())]
@@ -152,6 +155,61 @@ class Scan(LogicalPlan):
         )
 
 
+class FileWrite(UnaryNode):
+    def __init__(
+        self,
+        input: LogicalPlan,
+        root_dir: str,
+        storage_type: StorageType,
+        partition_cols: Optional[ExpressionList] = None,
+        compression: Optional[str] = None,
+    ) -> None:
+        assert (
+            storage_type == StorageType.PARQUET or storage_type == StorageType.CSV
+        ), "only parquet and csv is supported currently"
+        self._storage_type = storage_type
+        self._root_dir = root_dir
+        self._compression = compression
+        self._partition_cols = partition_cols
+        if self._partition_cols is not None:
+            self._partition_cols = self._partition_cols.resolve(input.schema())
+
+        for expr in input.schema():
+            assert ExpressionType.is_primitive(
+                expr.resolved_type()
+            ), f"we can currently only write out primitive types, got: {expr}"
+
+        schema = ExpressionList([ColumnExpression("file_path", ExpressionType.from_py_type(str))])
+        schema = schema.resolve()
+
+        super().__init__(schema, input.partition_spec(), op_level=OpLevel.PARTITION)
+        self._register_child(input)
+
+    def __repr__(self) -> str:
+        return f"FileWrite\n\toutput={self.schema()}"
+
+    def required_columns(self) -> ExpressionList:
+        raise NotImplementedError()
+
+    def resource_request(self) -> ResourceRequest:
+        return ResourceRequest.default()
+
+    def _local_eq(self, other: Any) -> bool:
+        return (
+            isinstance(other, FileWrite)
+            and self.schema() == other.schema()
+            and self._storage_type == other._storage_type
+            and self._root_dir == other._root_dir
+            and self._compression == other._compression
+        )
+
+    def rebuild(self) -> LogicalPlan:
+        raise NotImplementedError("We can not rebuild a filewrite due to side effects")
+
+    def copy_with_new_input(self, new_input: UnaryNode) -> UnaryNode:
+        raise NotImplementedError()
+
+
 class Filter(UnaryNode):
     """Which rows to keep"""
 
@@ -163,7 +221,7 @@ class Filter(UnaryNode):
         self._predicate = predicate.resolve(input.schema())
 
         resolved_type = self._predicate.exprs[0].resolved_type()
-        if not ExpressionType.is_logical(resolved_type):
+        if resolved_type != ExpressionType.logical():
             raise ValueError(
                 f"Expected expression {self._predicate.exprs[0]} to resolve to type LOGICAL, but received: {resolved_type}"
             )
@@ -218,22 +276,26 @@ class Projection(UnaryNode):
 
 
 class Sort(UnaryNode):
-    def __init__(self, input: LogicalPlan, sort_by: ExpressionList, desc: bool = False) -> None:
+    def __init__(
+        self, input: LogicalPlan, sort_by: ExpressionList, descending: Union[List[bool], bool] = False
+    ) -> None:
         pspec = PartitionSpec(scheme=PartitionScheme.RANGE, num_partitions=input.num_partitions(), by=sort_by)
         super().__init__(input.schema().to_column_expressions(), partition_spec=pspec, op_level=OpLevel.GLOBAL)
         self._register_child(input)
-        assert len(sort_by.exprs) == 1, "we can only sort with 1 expression"
         self._sort_by = sort_by.resolve(input_schema=input.schema())
-        self._desc = desc
+        if isinstance(descending, bool):
+            self._descending = [descending for _ in self._sort_by]
+        else:
+            self._descending = descending
 
     def __repr__(self) -> str:
-        return f"Sort\n\toutput={self.schema()}\n\tsort_by={self._sort_by}\n\tdesc={self._desc}"
+        return f"Sort\n\toutput={self.schema()}\n\tsort_by={self._sort_by}\n\tdesc={self._descending}"
 
     def resource_request(self) -> ResourceRequest:
         return self._sort_by.resource_request()
 
     def copy_with_new_input(self, new_input: LogicalPlan) -> Sort:
-        return Sort(new_input, sort_by=self._sort_by, desc=self._desc)
+        return Sort(new_input, sort_by=self._sort_by, descending=self._descending)
 
     def required_columns(self) -> ExpressionList:
         return self._sort_by.required_columns()
@@ -243,11 +305,65 @@ class Sort(UnaryNode):
             isinstance(other, Sort)
             and self.schema() == other.schema()
             and self._sort_by == self._sort_by
-            and self._desc == self._desc
+            and self._descending == self._descending
         )
 
     def rebuild(self) -> LogicalPlan:
-        return Sort(input=self._children()[0].rebuild(), sort_by=self._sort_by.unresolve(), desc=self._desc)
+        return Sort(input=self._children()[0].rebuild(), sort_by=self._sort_by.unresolve(), descending=self._descending)
+
+
+TMapPartitionOp = TypeVar("TMapPartitionOp", bound=MapPartitionOp)
+
+
+class MapPartition(UnaryNode, Generic[TMapPartitionOp]):
+    def __init__(self, input: LogicalPlan, map_partition_op: TMapPartitionOp) -> None:
+        self._map_partition_op = map_partition_op
+        super().__init__(
+            self._map_partition_op.get_output_schema(input.schema()),
+            partition_spec=input.partition_spec(),
+            op_level=OpLevel.PARTITION,
+        )
+        self._register_child(input)
+
+    def __repr__(self) -> str:
+        return f"MapPartition\n\top={self._map_partition_op}"
+
+    def resource_request(self) -> ResourceRequest:
+        return self._map_partition_op.resource_request()
+
+    def _local_eq(self, other: Any) -> bool:
+        return (
+            isinstance(other, MapPartition)
+            and self.schema() == other.schema()
+            and self._map_partition_op == other._map_partition_op
+        )
+
+    def eval_partition(self, partition: vPartition) -> vPartition:
+        return self._map_partition_op.run(partition)
+
+
+class Explode(MapPartition[ExplodeOp]):
+    def __init__(self, input: LogicalPlan, explode_expressions: ExpressionList):
+        explode_expressions_resolved = ExpressionList([e._explode() for e in explode_expressions]).resolve(
+            input.schema()
+        )
+        map_partition_op = ExplodeOp(explode_columns=explode_expressions_resolved)
+        super().__init__(
+            input,
+            map_partition_op,
+        )
+
+    def required_columns(self) -> ExpressionList:
+        return self._map_partition_op.explode_columns.required_columns()
+
+    def rebuild(self) -> LogicalPlan:
+        return Explode(
+            self._children()[0].rebuild(),
+            self._map_partition_op.explode_columns.unresolve(),
+        )
+
+    def copy_with_new_input(self, new_input: UnaryNode) -> Explode:
+        return Explode(new_input, explode_expressions=self._map_partition_op.explode_columns)
 
 
 class LocalLimit(UnaryNode):
@@ -318,12 +434,15 @@ class Repartition(UnaryNode):
     def __init__(
         self, input: LogicalPlan, partition_by: ExpressionList, num_partitions: int, scheme: PartitionScheme
     ) -> None:
+        resolved_part_by = partition_by.resolve(input.schema().to_column_expressions())
         pspec = PartitionSpec(
-            scheme=scheme, num_partitions=num_partitions, by=partition_by if len(partition_by) > 0 else None
+            scheme=scheme,
+            num_partitions=num_partitions,
+            by=resolved_part_by.to_column_expressions() if len(partition_by) > 0 else None,
         )
         super().__init__(input.schema().to_column_expressions(), partition_spec=pspec, op_level=OpLevel.GLOBAL)
         self._register_child(input)
-        self._partition_by = partition_by.resolve(self.schema())
+        self._partition_by = resolved_part_by
         self._scheme = scheme
         if scheme == PartitionScheme.RANDOM and len(partition_by.names) > 0:
             raise ValueError("Can not pass in random partitioning and partition_by args")
@@ -554,7 +673,7 @@ class Join(BinaryNode):
             num_partitions = min(left.num_partitions(), right.num_partitions())
             right_id_set = self._right_on.to_id_set()
             filtered_right = [e for e in right.schema() if e.get_id() not in right_id_set]
-            schema = left.schema().union(ExpressionList(filtered_right), strict=False, rename_dup="right.")
+            schema = left.schema().union(ExpressionList(filtered_right), rename_dup="right.")
 
         left = Repartition(left, partition_by=self._left_on, num_partitions=num_partitions, scheme=PartitionScheme.HASH)
         right = Repartition(
