@@ -1,6 +1,7 @@
 from __future__ import annotations
 
 import collections
+import math
 import operator
 import random
 from abc import abstractmethod
@@ -10,6 +11,7 @@ from typing import (
     Callable,
     ClassVar,
     Generic,
+    Iterable,
     Iterator,
     List,
     Optional,
@@ -18,16 +20,21 @@ from typing import (
     Type,
     TypeVar,
     Union,
+    cast,
 )
 
 import numpy as np
-import pandas as pd
+import polars as pl
 import pyarrow as pa
 import pyarrow.compute as pac
 from pandas.core.reshape.merge import get_join_indexers
 
 from daft.execution.operators import OperatorEnum, OperatorEvaluator
 from daft.internal.hashing import hash_chunked_array
+from daft.internal.kernels.search_sorted import search_sorted
+
+# A type representing some Python scalar (non-series/array like object)
+PyScalar = Any
 
 ArrType = TypeVar("ArrType", bound=collections.abc.Sequence)
 UnaryFuncType = Callable[[ArrType], ArrType]
@@ -99,16 +106,33 @@ class DataBlock(Generic[ArrType]):
         raise NotImplementedError()
 
     @abstractmethod
-    def to_numpy(self):
+    def to_numpy(self) -> np.ndarray:
+        raise NotImplementedError()
+
+    @abstractmethod
+    def to_polars(self) -> Union[PyScalar, pl.Series]:
+        raise NotImplementedError()
+
+    @abstractmethod
+    def to_arrow(self) -> pa.ChunkedArray:
         raise NotImplementedError()
 
     @classmethod
     def make_block(cls, data: Any) -> DataBlock:
         # Data is a sequence of data
+        # Arrow data: nested types get casted to a list and stored as PyListDataBlock
         if isinstance(data, pa.ChunkedArray):
+            if pa.types.is_nested(data.type):
+                return PyListDataBlock(data=data.to_pylist())
             return ArrowDataBlock(data=data)
         elif isinstance(data, pa.Array):
+            if pa.types.is_nested(data.type):
+                return PyListDataBlock(data=data.to_pylist())
             return ArrowDataBlock(data=pa.chunked_array([data]))
+        # Lists: always stored as a PyListDataBlock
+        elif isinstance(data, list):
+            return PyListDataBlock(data=data)
+        # Numpy data: attempt to cast to Arrow, nested types get casted to a list and stored as PyListDataBlock
         elif isinstance(data, np.ndarray):
             if data.dtype == np.object_ or len(data.shape) > 1:
                 try:
@@ -120,27 +144,10 @@ class DataBlock(Generic[ArrType]):
                 return ArrowDataBlock(data=pa.chunked_array([pa.array(data, type=arrow_type)]))
             arrow_type = pa.from_numpy_dtype(data.dtype)
             return ArrowDataBlock(data=pa.chunked_array([pa.array(data, type=arrow_type)]))
-        elif isinstance(data, pd.Series):
-            if data.dtype == np.object_:
-                try:
-                    arrow_type = pa.infer_type(data)
-                except pa.lib.ArrowInvalid:
-                    arrow_type = None
-                if arrow_type is None or pa.types.is_nested(arrow_type):
-                    return PyListDataBlock(data=data.to_list())
-                return ArrowDataBlock(data=pa.chunked_array([pa.array(data, type=arrow_type)]))
-            arrow_type = pa.Schema.from_pandas(pd.DataFrame({"0": data}))[0].type
-            return ArrowDataBlock(data=pa.chunked_array([pa.Array.from_pandas(data, type=arrow_type)]))
-        elif isinstance(data, list):
-            try:
-                arrow_type = pa.infer_type(data)
-            except pa.lib.ArrowInvalid:
-                arrow_type = None
-            if arrow_type is None or pa.types.is_nested(arrow_type):
-                return PyListDataBlock(data=data)
-            return ArrowDataBlock(data=pa.chunked_array([pa.array(data, type=arrow_type)]))
         # Data is a scalar
         elif isinstance(data, pa.Scalar):
+            if pa.types.is_nested(data.type):
+                return PyListDataBlock(data=data.as_py())
             return ArrowDataBlock(data=data)
         else:
             try:
@@ -188,29 +195,26 @@ class DataBlock(Generic[ArrType]):
         fn: Callable[[DataBlock[ArrType], DataBlock[ArrType]], DataBlock[ArrType]] = getattr(self.evaluator, op_name)
         return fn(self, other)
 
-    def partition(self, num: int, targets: DataBlock[ArrowArrType]) -> List[DataBlock[ArrType]]:
+    def partition(
+        self, num: int, pivots: np.ndarray, target_partitions: np.ndarray, argsorted_targets: DataBlock[ArrowArrType]
+    ) -> List[DataBlock[ArrType]]:
         assert not self.is_scalar(), "Cannot partition scalar DataBlock"
 
         # We first argsort the targets to group the same partitions together
-        argsort_indices = targets.argsort()
+        argsort_indices = argsorted_targets
 
         # We now perform a gather to make items targeting the same partition together
         reordered = self.take(argsort_indices)
-        sorted_targets = targets.take(argsort_indices)
-
-        sorted_targets_np = sorted_targets.data.to_numpy()
-        pivots = np.where(np.diff(sorted_targets_np, prepend=np.nan))[0]
 
         # We now split in the num partitions
         unmatched_partitions = reordered._split(pivots)
-        target_partitions = sorted_targets_np[pivots]
-
         target_partition_idx_to_match_idx = {target_idx: idx for idx, target_idx in enumerate(target_partitions)}
+        empty = self._make_empty()
 
         return [
             DataBlock.make_block(unmatched_partitions[target_partition_idx_to_match_idx[i]])
             if i in target_partition_idx_to_match_idx
-            else self._make_empty()
+            else empty
             for i in range(num)
         ]
 
@@ -250,13 +254,44 @@ class DataBlock(Generic[ArrType]):
     def sample(self, num: int) -> DataBlock[ArrType]:
         raise NotImplementedError()
 
+    @staticmethod
     @abstractmethod
-    def _argsort(self, desc: bool = False) -> DataBlock[ArrowArrType]:
+    def _argsort(blocks: List[DataBlock[ArrType]], descending: Optional[List[bool]] = None) -> DataBlock[ArrowArrType]:
         raise NotImplementedError()
 
-    def argsort(self, desc: bool = False) -> DataBlock[ArrowArrType]:
-        assert not self.is_scalar(), "Cannot sort scalar DataBlock"
-        return self._argsort(desc=desc)
+    @classmethod
+    def argsort(cls, blocks: List[DataBlock], descending: Optional[List[bool]] = None) -> DataBlock[ArrowArrType]:
+        first_type = type(blocks[0])
+        assert all(type(b) == first_type for b in blocks), "all block types must match"
+        size = len(blocks)
+        if descending is None:
+            descending = [False for _ in range(size)]
+        else:
+            assert size == len(descending), f"mismatch of size of descending and blocks {size} vs {len(descending)}"
+        return first_type._argsort(blocks, descending=descending)
+
+    @staticmethod
+    @abstractmethod
+    def _search_sorted(
+        sorted_blocks: List[DataBlock], keys: List[DataBlock], input_reversed: Optional[List[bool]] = None
+    ) -> DataBlock[ArrowArrType]:
+        raise NotImplementedError()
+
+    @classmethod
+    def search_sorted(
+        cls, sorted_blocks: List[DataBlock], keys: List[DataBlock], input_reversed: Optional[List[bool]] = None
+    ) -> DataBlock[ArrowArrType]:
+        first_type = type(sorted_blocks[0])
+        assert all(type(b) == first_type for b in sorted_blocks), "all block types must match"
+        assert all(type(b) == first_type for b in keys), "all block types must match"
+
+        size = len(sorted_blocks)
+        if input_reversed is None:
+            input_reversed = [False for _ in range(size)]
+        assert len(sorted_blocks) == len(keys)
+        assert size == len(input_reversed), f"mismatch of size of descending and blocks {size} vs {len(input_reversed)}"
+
+        return first_type._search_sorted(sorted_blocks=sorted_blocks, keys=keys, input_reversed=input_reversed)
 
     @staticmethod
     @abstractmethod
@@ -270,14 +305,6 @@ class DataBlock(Generic[ArrType]):
         assert all(type(b) == first_type for b in blocks), "all block types must match"
         return first_type._merge_blocks(blocks)
 
-    @abstractmethod
-    def search_sorted(self, pivots: DataBlock[ArrType], reverse: bool = False) -> DataBlock[ArrType]:
-        raise NotImplementedError()
-
-    @abstractmethod
-    def quantiles(self, num: int) -> DataBlock[ArrType]:
-        raise NotImplementedError()
-
     @abstractmethod
     def array_hash(self, seed: Optional[DataBlock[ArrType]] = None) -> DataBlock[ArrowArrType]:
         raise NotImplementedError()
@@ -344,6 +371,18 @@ class DataBlock(Generic[ArrType]):
 
         return to_rtn
 
+    @abstractmethod
+    def list_explode(self) -> Tuple[DataBlock[ArrType], DataBlock[ArrowArrType]]:
+        """Treats each row as a list and flattens the nested lists. Will throw an error if
+        elements in the block cannot be treated as a list.
+
+        Note: empty lists are flattened into a single `Null` element!
+
+        Returns a tuple `(exploded_block, list_lengths)` where `exploded_block` is the new block with the
+        topmost level of lists flattened, and `list_lengths` is the length of each list element before flattening.
+        """
+        raise NotImplementedError()
+
 
 T = TypeVar("T")
 
@@ -368,6 +407,14 @@ class PyListDataBlock(DataBlock[List[T]]):
             res[i] = obj
         return res
 
+    def to_polars(self) -> Union[PyScalar, pl.Series]:
+        if self.is_scalar():
+            return self.data
+        return pl.Series(self.data, dtype=pl.datatypes.Object)
+
+    def to_arrow(self) -> pa.ChunkedArray:
+        raise NotImplementedError("can not convert pylist block to arrow")
+
     def _filter(self, mask: DataBlock[ArrowArrType]) -> DataBlock[List[T]]:
         return PyListDataBlock(data=[item for keep, item in zip(mask.iter_py(), self.data) if keep])
 
@@ -377,7 +424,14 @@ class PyListDataBlock(DataBlock[List[T]]):
     def sample(self, num: int) -> DataBlock[List[T]]:
         return PyListDataBlock(data=random.sample(self.data, num))
 
-    def _argsort(self, desc: bool = False) -> DataBlock[ArrowArrType]:
+    @staticmethod
+    def _search_sorted(
+        sorted_blocks: List[DataBlock], keys: List[DataBlock], input_reversed: Optional[List[bool]] = None
+    ) -> DataBlock[ArrowArrType]:
+        raise NotImplementedError()
+
+    @staticmethod
+    def _argsort(blocks: List[DataBlock[ArrType]], descending: Optional[List[bool]] = None) -> DataBlock[ArrowArrType]:
         raise NotImplementedError("Sorting by Python objects is not implemented")
 
     def _make_empty(self) -> DataBlock[List[T]]:
@@ -393,12 +447,6 @@ class PyListDataBlock(DataBlock[List[T]]):
             concatted_data.extend(block.data)
         return PyListDataBlock(data=concatted_data)
 
-    def search_sorted(self, pivots: DataBlock[List[T]], reverse: bool = False) -> DataBlock[ArrowArrType]:
-        raise NotImplementedError("Sorting by Python objects is not implemented")
-
-    def quantiles(self, num: int) -> DataBlock[List[T]]:
-        raise NotImplementedError("Sorting by Python objects is not implemented")
-
     def array_hash(self, seed: Optional[DataBlock[ArrType]] = None) -> DataBlock[ArrowArrType]:
         # TODO(jay): seed is ignored here, but perhaps we need to set it in PYTHONHASHSEED?
         hashes = [hash(x) for x in self.data]
@@ -419,11 +467,37 @@ class PyListDataBlock(DataBlock[List[T]]):
     ) -> Tuple[DataBlock[List[T]], DataBlock[List[T]]]:
         raise NotImplementedError()
 
+    def list_explode(self) -> Tuple[PyListDataBlock, DataBlock[ArrowArrType]]:
+        lengths = []
+        exploded = []
+        for item in self.data:
+            length = 0
+            if item is not None:
+                for nested_element in cast(Iterable, item):
+                    length += 1
+                    exploded.append(nested_element)
+            # empty lists or Null elements explode into a `Null`
+            if length == 0:
+                exploded.append(None)
+                lengths.append(1)
+            else:
+                lengths.append(length)
+        return (
+            PyListDataBlock(exploded),
+            DataBlock.make_block(pa.array(lengths)),
+        )
+
 
 ArrowArrType = Union[pa.ChunkedArray, pa.Scalar]
 
 
 class ArrowDataBlock(DataBlock[ArrowArrType]):
+    def __init__(self, data: ArrowArrType) -> None:
+        assert not pa.types.is_nested(
+            data.type
+        ), f"Cannot create ArrowDataBlock with nested type: {data.type}, please report this to Daft developers."
+        super().__init__(data)
+
     def __reduce__(self) -> Tuple:
         if len(self.data) == 0:
             return ArrowDataBlock, (self._make_empty().data,)
@@ -435,6 +509,14 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
             return self.data.as_py()
         return self.data.to_numpy()
 
+    def to_polars(self) -> Union[PyScalar, pl.Series]:
+        if isinstance(self.data, pa.Scalar):
+            return self.data.as_py()
+        return pl.Series(self.data)
+
+    def to_arrow(self) -> pa.ChunkedArray:
+        return self.data
+
     def is_scalar(self) -> bool:
         return isinstance(self.data, pa.Scalar)
 
@@ -443,53 +525,86 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
             py_scalar = self.data.as_py()
             while True:
                 yield py_scalar
-        yield from self.data.to_numpy()
+        yield from self.data.to_pylist()
 
-    def _argsort(self, desc: bool = False) -> DataBlock[ArrowArrType]:
-        order = "descending" if desc else "ascending"
-        sort_indices = pac.array_sort_indices(self.data, order=order)
-        return ArrowDataBlock(data=sort_indices)
+    # def _argsort(self, desc: bool = False) -> DataBlock[ArrowArrType]:
+    #     order = "descending" if desc else "ascending"
+    #     sort_indices = pac.array_sort_indices(self.data, order=order)
+    #     return ArrowDataBlock(data=sort_indices)
+
+    @staticmethod
+    def _argsort(blocks: List[DataBlock[ArrType]], descending: Optional[List[bool]] = None) -> DataBlock[ArrowArrType]:
+        arrs = [a.data for a in blocks]
+        arr_names = [f"a_{i}" for i in range(len(blocks))]
+        if descending is None:
+            descending = [False for _ in range(len(blocks))]
+        order = ["descending" if desc else "ascending" for desc in descending]
+        table = pa.table(arrs, names=arr_names)
+        indices = pac.sort_indices(table, sort_keys=list(zip(arr_names, order)))
+        return DataBlock.make_block(indices)
+
+    @staticmethod
+    def _search_sorted(
+        sorted_blocks: List[DataBlock], keys: List[DataBlock], input_reversed: Optional[List[bool]] = None
+    ) -> DataBlock[ArrowArrType]:
+
+        arr_names = [f"a_{i}" for i in range(len(sorted_blocks))]
+        if input_reversed is None:
+            input_reversed = [False for _ in range(len(sorted_blocks))]
+        sorted_table = pa.table([a.data for a in sorted_blocks], names=arr_names)
+        key_table = pa.table([a.data for a in keys], names=arr_names)
+        assert sorted_table.schema == key_table.schema
+        result = search_sorted(sorted_table, key_table, input_reversed=input_reversed)
+        return DataBlock.make_block(result)
 
     def _filter(self, mask: DataBlock[ArrowArrType]) -> DataBlock[ArrowArrType]:
         return self._binary_op(mask, fn=pac.array_filter)
 
     def _take(self, indices: DataBlock[ArrowArrType]) -> DataBlock[ArrowArrType]:
+
+        if pa.types.is_binary(self.data.type) or pa.types.is_string(self.data.type):
+            bytes_per_row = self.data.nbytes / max(len(self.data), 1)
+            expected_size = bytes_per_row * len(indices) * 2  # with safety factor
+            if expected_size >= (2**31):
+                # if we have a binary or string array thats result will be is larger than 2GB
+                # then we need to chunk the indice array so the result array will also be chunk
+                # this is a hack since arrow's kernels don't correctly chunk with the arraybuilder
+                # for take.
+                num_chunks = math.ceil(expected_size / (2**31))
+                chunk_size = len(indices) // num_chunks
+                new_indices_subarray = []
+                for i in range(num_chunks):
+                    new_indices_subarray.extend(indices.data.slice(i * chunk_size, chunk_size).chunks)
+                indices = DataBlock.make_block(pa.chunked_array(new_indices_subarray))
+
         return self._binary_op(indices, fn=partial(pac.take, boundscheck=False))
 
     def _split(self, pivots: np.ndarray) -> Sequence[ArrowArrType]:
-        splitted: Sequence[np.ndarray] = np.split(self.data, pivots)[1:]
-        return splitted
+        to_return = []
+        for i in range(len(pivots) - 1):
+            offset = pivots[i]
+            size = pivots[i + 1] - offset
+            # Combine chunks otherwise arrow has a serialization issue with giant memory
+            to_return.append(self.data.slice(offset, size).combine_chunks())
+        if len(pivots) > 0:
+            to_return.append(self.data.slice(pivots[-1]).combine_chunks())
+        return to_return
 
     @staticmethod
     def _merge_blocks(blocks: List[DataBlock[ArrowArrType]]) -> DataBlock[ArrowArrType]:
         all_chunks = []
         for block in blocks:
             all_chunks.extend(block.data.chunks)
-        return ArrowDataBlock(data=pa.chunked_array(all_chunks))
+        return DataBlock.make_block(pa.chunked_array(all_chunks))
 
     def _make_empty(self) -> DataBlock[ArrowArrType]:
-        return ArrowDataBlock(data=pa.chunked_array([[]], type=self.data.type))
+        return DataBlock.make_block(data=pa.chunked_array([[]], type=self.data.type))
 
     def sample(self, num: int) -> DataBlock[ArrowArrType]:
         if num >= len(self):
             return self
         sampled = np.random.choice(self.data, num, replace=False)
-        return ArrowDataBlock(data=pa.chunked_array([sampled]))
-
-    def search_sorted(self, pivots: DataBlock, reverse: bool = False) -> DataBlock[ArrowArrType]:
-        arr = self.data.to_numpy()
-        indices = np.searchsorted(pivots.data.to_numpy(), arr)
-        if reverse:
-            indices = len(pivots) - indices
-        return ArrowDataBlock(data=pa.chunked_array([indices]))
-
-    def quantiles(self, num: int) -> DataBlock[ArrowArrType]:
-        quantiles = np.linspace(1.0 / num, 1.0, num)[:-1]
-        if NUMPY_MINOR_VERSION < 22:
-            pivots = np.quantile(self.data.to_numpy(), quantiles, interpolation="nearest")
-        else:
-            pivots = np.quantile(self.data.to_numpy(), quantiles, method="closest_observation")
-        return DataBlock.make_block(data=pivots)
+        return DataBlock.make_block(data=pa.chunked_array([sampled]))
 
     def array_hash(self, seed: Optional[DataBlock[ArrowArrType]] = None) -> DataBlock[ArrowArrType]:
         assert isinstance(self.data, pa.ChunkedArray)
@@ -507,7 +622,7 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
             raise TypeError(f"cannot hash {pa_type}")
         hashed = hash_chunked_array(data_to_hash)
         if seed is None:
-            return ArrowDataBlock(data=hashed)
+            return DataBlock.make_block(data=hashed)
         else:
             seed_pa_type = seed.data.type
             if not (pa.types.is_uint64(seed_pa_type)):
@@ -515,7 +630,7 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
 
             seed_arr = seed.data.to_numpy()
             seed_arr = seed_arr ^ (hashed.to_numpy() + 0x9E3779B9 + (seed_arr << 6) + (seed_arr >> 2))
-            return ArrowDataBlock(data=pa.chunked_array([seed_arr]))
+            return DataBlock.make_block(data=pa.chunked_array([seed_arr]))
 
     def agg(self, op: str) -> DataBlock[ArrowArrType]:
 
@@ -539,14 +654,44 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
     def _group_by_agg(
         group_by: List[DataBlock[ArrowArrType]], to_agg: List[DataBlock[ArrowArrType]], agg_ops: List[str]
     ) -> Tuple[List[DataBlock[ArrowArrType]], List[DataBlock[ArrowArrType]]]:
-        arrs = [a.data for a in group_by]
-        arrs.extend([a.data for a in to_agg])
+        group_arrs = [a.data for a in group_by]
+        agg_arrs = [a.data for a in to_agg]
+        arrs = group_arrs + agg_arrs
         group_names = [f"g_{i}" for i in range(len(group_by))]
         agg_names = [f"a_{i}" for i in range(len(to_agg))]
-        table = pa.table(arrs, names=group_names + agg_names).combine_chunks()
-        agged = table.group_by(group_names).aggregate([(a_name, op) for a_name, op in zip(agg_names, agg_ops)])
-        gcols: List[DataBlock] = [ArrowDataBlock(agged[g_name]) for g_name in group_names]
-        acols: List[DataBlock] = [ArrowDataBlock(agged[f"{a_name}_{op}"]) for a_name, op in zip(agg_names, agg_ops)]
+        table = pa.table(arrs, names=group_names + agg_names)
+        pl_table = pl.from_arrow(table, rechunk=True)
+
+        exprs = []
+        grouped_expected_arrow_type = [a.type for a in group_arrs]
+        agg_expected_arrow_type = []
+        for an, op, arr in zip(agg_names, agg_ops, agg_arrs):
+            if op == "sum":
+                exprs.append(pl.sum(an))
+                agg_expected_arrow_type.append(arr.type)
+            elif op == "mean":
+                exprs.append(pl.mean(an))
+                agg_expected_arrow_type.append(pa.float64())
+            elif op == "min":
+                exprs.append(pl.min(an))
+                agg_expected_arrow_type.append(arr.type)
+            elif op == "max":
+                exprs.append(pl.max(an))
+                agg_expected_arrow_type.append(arr.type)
+            elif op == "count":
+                exprs.append(pl.count(an))
+                agg_expected_arrow_type.append(pa.int64())
+            else:
+                raise NotImplementedError()
+
+        pl_agged = pl_table.groupby(group_names).agg(exprs)
+        agged = pl_agged.to_arrow()
+        gcols: List[DataBlock] = [
+            DataBlock.make_block(agged[g_name].cast(t)) for g_name, t in zip(group_names, grouped_expected_arrow_type)
+        ]
+        acols: List[DataBlock] = [
+            DataBlock.make_block(agged[f"{a_name}"].cast(t)) for a_name, t in zip(agg_names, agg_expected_arrow_type)
+        ]
         return gcols, acols
 
     @staticmethod
@@ -562,25 +707,50 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
 
         return DataBlock.make_block(left_index), DataBlock.make_block(right_index)
 
+    def list_explode(self) -> Tuple[DataBlock[ArrType], DataBlock[ArrowArrType]]:
+        """Treats each row as a list and flattens the nested lists. Will throw an error if
+        elements in the block cannot be treated as a list.
+        """
+        data = self.data
+        data = pac.if_else(pac.equal(pac.list_value_length(data), pa.scalar(0)), pa.scalar([None], data.type), data)
+        data = pac.if_else(data.is_null(), pa.scalar([None], data.type), data)
+        exploded = pac.list_flatten(data)
+        list_lengths = pac.list_value_length(data)
+        return (
+            DataBlock.make_block(exploded),
+            DataBlock.make_block(list_lengths),
+        )
+
 
 def arrow_mod(arr, m):
-    return np.mod(arr, m.as_py())
+    if isinstance(m, pa.Scalar):
+        return np.mod(arr, m.as_py())
+    return np.mod(arr, m)
 
 
 def arrow_floordiv(arr, m):
-    return pac.floor(pac.divide(arr, m))
+    floordiv = pac.floor(pac.divide(arr, m))
+    if pa.types.is_integer(arr.type) and pa.types.is_integer(m.type):
+        floordiv = floordiv.cast(pa.int64())
+    return floordiv
 
 
 def arrow_str_contains(arr: pa.ChunkedArray, pattern: pa.StringScalar):
+    if pa.types.is_null(arr.type):
+        return arr
     substring_counts = pac.count_substring(arr, pattern=pattern.as_py())
     return pac.not_equal(substring_counts, pa.scalar(0))
 
 
 def arrow_str_endswith(arr: pa.ChunkedArray, pattern: pa.StringScalar):
+    if pa.types.is_null(arr.type):
+        return arr
     return pac.ends_with(arr, pattern=pattern.as_py())
 
 
 def arrow_str_startswith(arr: pa.ChunkedArray, pattern: pa.StringScalar):
+    if pa.types.is_null(arr.type):
+        return arr
     return pac.starts_with(arr, pattern=pattern.as_py())
 
 
@@ -606,11 +776,6 @@ class ArrowEvaluator(OperatorEvaluator["ArrowDataBlock"]):
     NEGATE = _arr_unary_op(pac.negate)
     POSITIVE = ArrowDataBlock.identity
     ABS = _arr_unary_op(pac.abs)
-    SUM = ArrowDataBlock.identity
-    MEAN = ArrowDataBlock.identity
-    MIN = ArrowDataBlock.identity
-    MAX = ArrowDataBlock.identity
-    COUNT = ArrowDataBlock.identity
     INVERT = _arr_unary_op(pac.invert)
     ADD = _arr_bin_op(pac.add)
     SUB = _arr_bin_op(pac.subtract)
@@ -637,8 +802,18 @@ class ArrowEvaluator(OperatorEvaluator["ArrowDataBlock"]):
     DT_DAY_OF_WEEK = _arr_unary_op(pac.day_of_week)
     IS_NULL = _arr_unary_op(pac.is_null)
     IS_NAN = _arr_unary_op(pac.is_nan)
+    EXPLODE = _arr_unary_op(pac.list_flatten)
     IF_ELSE = _arr_ternary_op(pac.if_else)
 
+    # Placeholders: these change the cardinality of the block and should never be evaluated from the Evaluator
+    # They exist on the Evaluator only to provide correct typing information
+    SUM = ArrowDataBlock.identity
+    MEAN = ArrowDataBlock.identity
+    MIN = ArrowDataBlock.identity
+    MAX = ArrowDataBlock.identity
+    COUNT = ArrowDataBlock.identity
+    EXPLODE = ArrowDataBlock.identity
+
 
 ArrowDataBlock.evaluator = ArrowEvaluator
 
@@ -647,18 +822,18 @@ IN_2 = TypeVar("IN_2")
 OUT = TypeVar("OUT")
 
 
-def make_map_unary(f: Callable[[IN_1], OUT]) -> Callable[[DataBlock[Sequence[IN_1]]], DataBlock[Sequence[OUT]]]:
-    def map_f(values: DataBlock[Sequence[IN_1]]) -> DataBlock[Sequence[OUT]]:
-        return DataBlock.make_block(list(map(f, (tup[0] for tup in zip_blocks_as_py(values)))))
+def make_map_unary(f: Callable[[IN_1], OUT]) -> Callable[[DataBlock[Sequence[IN_1]]], PyListDataBlock[OUT]]:
+    def map_f(values: DataBlock[Sequence[IN_1]]) -> PyListDataBlock[OUT]:
+        return PyListDataBlock(list(map(f, (tup[0] for tup in zip_blocks_as_py(values)))))
 
     return map_f
 
 
 def make_map_binary(
     f: Callable[[IN_1, IN_2], OUT]
-) -> Callable[[DataBlock[Sequence[IN_1]], DataBlock[Sequence[IN_2]]], DataBlock[Sequence[OUT]]]:
-    def map_f(values: DataBlock[Sequence[IN_1]], others: DataBlock[Sequence[IN_2]]) -> DataBlock[Sequence[OUT]]:
-        return DataBlock.make_block(list(f(v, o) for v, o in zip_blocks_as_py(values, others)))
+) -> Callable[[DataBlock[Sequence[IN_1]], DataBlock[Sequence[IN_2]]], PyListDataBlock[OUT]]:
+    def map_f(values: DataBlock[Sequence[IN_1]], others: DataBlock[Sequence[IN_2]]) -> PyListDataBlock[OUT]:
+        return PyListDataBlock(list(f(v, o) for v, o in zip_blocks_as_py(values, others)))
 
     return map_f
 
@@ -673,19 +848,14 @@ def pylist_is_none(obj: Any):
 
 def pylist_if_else(
     cond: DataBlock[Sequence[IN_1]], x: DataBlock[Sequence[IN_2]], y: DataBlock[Sequence[IN_2]]
-) -> DataBlock[Sequence[IN_2]]:
-    return DataBlock.make_block([xitem if c else yitem for c, xitem, yitem in zip_blocks_as_py(cond, x, y)])
+) -> PyListDataBlock[IN_2]:
+    return PyListDataBlock([xitem if c else yitem for c, xitem, yitem in zip_blocks_as_py(cond, x, y)])
 
 
 class PyListEvaluator(OperatorEvaluator["PyListDataBlock"]):
     NEGATE = make_map_unary(operator.neg)
     POSITIVE = PyListDataBlock.identity
     ABS = make_map_unary(operator.abs)
-    SUM = PyListDataBlock.identity
-    MEAN = PyListDataBlock.identity
-    MIN = PyListDataBlock.identity
-    MAX = PyListDataBlock.identity
-    COUNT = PyListDataBlock.identity
     INVERT = make_map_unary(operator.invert)
     ADD = make_map_binary(operator.add)
     SUB = make_map_binary(operator.sub)
@@ -706,6 +876,15 @@ class PyListEvaluator(OperatorEvaluator["PyListDataBlock"]):
     IS_NULL = make_map_unary(pylist_is_none)
     IF_ELSE = pylist_if_else
 
+    # Placeholders: these change the cardinality of the block and should never be evaluated from the Evaluator
+    # They exist on the Evaluator only to provide correct typing information
+    SUM = PyListDataBlock.identity
+    MEAN = PyListDataBlock.identity
+    MIN = PyListDataBlock.identity
+    MAX = PyListDataBlock.identity
+    COUNT = PyListDataBlock.identity
+    EXPLODE = PyListDataBlock.identity
+
     # Unary operations that should never run on a PyListDataBlock because they are represented by
     # Arrow primitives and should always be housed in an ArrowDataBlock
     STR_CONTAINS = assert_invalid_pylist_operation
