@@ -4,17 +4,29 @@ import dataclasses
 from abc import abstractmethod
 from dataclasses import dataclass
 from functools import partial
-from typing import Any, Callable, Dict, Generic, List, Optional, Tuple, TypeVar
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Generic,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    TypeVar,
+)
+from uuid import uuid4
 
 import numpy as np
 import pandas as pd
 import pyarrow as pa
+from pyarrow import dataset as pada
 
-from daft.expressions import ColID, Expression, ExpressionExecutor
+from daft.execution.operators import OperatorEnum
+from daft.expressions import ColID, ColumnExpression, Expression, ExpressionExecutor
 from daft.logical.schema import ExpressionList
 from daft.runners.blocks import ArrowArrType, DataBlock, PyListDataBlock
-
-from ..execution.operators import OperatorEnum, PythonExpressionType
+from daft.types import ExpressionType
 
 PartID = int
 
@@ -32,9 +44,20 @@ class PyListTile:
     def apply(self, func: Callable[[DataBlock], DataBlock]) -> PyListTile:
         return dataclasses.replace(self, block=func(self.block))
 
-    def split_by_index(self, num_partitions: int, target_partition_indices: DataBlock) -> List[PyListTile]:
-        assert len(target_partition_indices) == len(self)
-        new_blocks = self.block.partition(num_partitions, targets=target_partition_indices)
+    def split_by_index(
+        self,
+        num_partitions: int,
+        pivots: np.ndarray,
+        target_partitions: np.ndarray,
+        argsorted_target_partition_indices: DataBlock,
+    ) -> List[PyListTile]:
+        assert len(argsorted_target_partition_indices) == len(self)
+        new_blocks = self.block.partition(
+            num_partitions,
+            pivots=pivots,
+            target_partitions=target_partitions,
+            argsorted_targets=argsorted_target_partition_indices,
+        )
         assert len(new_blocks) == num_partitions
         return [dataclasses.replace(self, block=nb, partition_id=i) for i, nb in enumerate(new_blocks)]
 
@@ -136,15 +159,15 @@ class vPartition:
         for col_expr in column_exprs:
             col_id = col_expr.get_id()
             col_name = col_expr.name()
-            arr = (
-                data[col_name]
-                if isinstance(col_expr.resolved_type(), PythonExpressionType)
-                else pa.array(data[col_expr.name()])
-            )
+            arr = data[col_name] if ExpressionType.is_py(col_expr.resolved_type()) else pa.array(data[col_expr.name()])
             block: DataBlock = DataBlock.make_block(arr)
             tiles[col_id] = PyListTile(column_id=col_id, column_name=col_name, partition_id=partition_id, block=block)
         return vPartition(columns=tiles, partition_id=partition_id)
 
+    def to_pydict(self) -> Dict[str, Sequence]:
+        output_schema = [(tile.column_name, id) for id, tile in self.columns.items()]
+        return {name: self.columns[id].block.data for name, id in output_schema}
+
     def to_pandas(self, schema: Optional[ExpressionList] = None) -> pd.DataFrame:
         if schema is not None:
             output_schema = [(expr.name(), expr.get_id()) for expr in schema]
@@ -180,10 +203,29 @@ class vPartition:
             mask = mask.run_binary_operator(to_and.block, OperatorEnum.AND)
         return self.for_each_column_block(partial(DataBlock.filter, mask=mask))
 
-    def sort(self, sort_key: Expression, desc: bool = False) -> vPartition:
-        sort_tile = self.eval_expression(sort_key)
-        argsort_idx = sort_tile.apply(partial(DataBlock.argsort, desc=desc))
-        return self.take(argsort_idx.block)
+    def argsort(self, sort_keys: ExpressionList, descending: Optional[List[bool]] = None) -> DataBlock:
+        sorted = self.eval_expression_list(sort_keys)
+        keys = list(sorted.columns.keys())
+
+        if descending is None:
+            descending = [False for _ in keys]
+
+        idx = DataBlock.argsort([sorted.columns[k].block for k in keys], descending=descending)
+        return idx
+
+    def sort(self, sort_keys: ExpressionList, descending: Optional[List[bool]] = None) -> vPartition:
+        idx = self.argsort(sort_keys=sort_keys, descending=descending)
+        return self.take(idx)
+
+    def search_sorted(self, keys: vPartition, input_reversed: Optional[List[bool]] = None) -> DataBlock:
+        assert self.columns.keys() == keys.columns.keys()
+        col_ids = list(self.columns.keys())
+        idx = DataBlock.search_sorted(
+            [self.columns[k].block for k in col_ids],
+            [keys.columns[k].block for k in col_ids],
+            input_reversed=input_reversed,
+        )
+        return idx
 
     def take(self, indices: DataBlock) -> vPartition:
         return self.for_each_column_block(partial(DataBlock.take, indices=indices))
@@ -229,15 +271,80 @@ class vPartition:
     def split_by_index(self, num_partitions: int, target_partition_indices: DataBlock) -> List[vPartition]:
         assert len(target_partition_indices) == len(self)
         new_partition_to_columns: List[Dict[ColID, PyListTile]] = [{} for _ in range(num_partitions)]
+        argsort_targets = DataBlock.argsort([target_partition_indices])
+        sorted_targets = target_partition_indices.take(argsort_targets)
+        sorted_targets_np = sorted_targets.to_numpy()
+        pivots = np.where(np.diff(sorted_targets_np, prepend=np.nan))[0]
+        target_partitions = sorted_targets_np[pivots]
+
         for col_id, tile in self.columns.items():
             new_tiles = tile.split_by_index(
-                num_partitions=num_partitions, target_partition_indices=target_partition_indices
+                num_partitions=num_partitions,
+                pivots=pivots,
+                target_partitions=target_partitions,
+                argsorted_target_partition_indices=argsort_targets,
             )
             for part_id, nt in enumerate(new_tiles):
                 new_partition_to_columns[part_id][col_id] = nt
 
         return [vPartition(partition_id=i, columns=columns) for i, columns in enumerate(new_partition_to_columns)]
 
+    def quantiles(self, num: int) -> vPartition:
+        self_size = len(self)
+        sample_idx_np = np.linspace(self_size / num, self_size, num)[:-1].round().astype(np.int32)
+        return self.take(DataBlock.make_block(sample_idx_np))
+
+    def explode(self, columns: ExpressionList) -> vPartition:
+        partition_to_explode = self.eval_expression_list(columns)
+        exploded_col_names = {tile.column_name for tile in partition_to_explode.columns.values()}
+        partition_to_repeat = vPartition(
+            {cid: tile for cid, tile in self.columns.items() if tile.column_name not in exploded_col_names},
+            partition_id=self.partition_id,
+        )
+
+        exploded_cols = {}
+        found_list_lengths = None
+        for col_id, tile in partition_to_explode.columns.items():
+            exploded_block, list_lengths = tile.block.list_explode()
+
+            # Ensure that each row has the same length as other rows
+            if found_list_lengths is None:
+                found_list_lengths = list_lengths
+            else:
+                if list_lengths != found_list_lengths:
+                    raise RuntimeError(
+                        ".explode expects columns to have the same length in each row, "
+                        "but found row(s) with mismatched lengths"
+                    )
+
+            exploded_cols[col_id] = PyListTile(
+                column_id=tile.column_id,
+                column_name=tile.column_name,
+                partition_id=tile.partition_id,
+                block=exploded_block,
+            )
+        assert found_list_lengths is not None, "At least one column must be specified to explode"
+
+        if len(found_list_lengths) == 0:
+            return vPartition(
+                {
+                    **exploded_cols,
+                    **partition_to_repeat.columns,
+                },
+                partition_id=self.partition_id,
+            )
+
+        # Use the `found_list_lengths` to generate an array of indices to take from other columns (e.g. [0, 0, 1, 1, 1, 2, ...])
+        list_length_cumsum = found_list_lengths.to_numpy().cumsum()
+        take_indices = np.zeros(list_length_cumsum[-1], dtype="int64")
+        take_indices[0] = 1
+        take_indices[list_length_cumsum[:-1]] = 1
+        take_indices = take_indices.cumsum()
+        take_indices = take_indices - 1
+
+        repeated_partition = partition_to_repeat.take(DataBlock.make_block(take_indices))
+        return vPartition({**exploded_cols, **repeated_partition.columns}, partition_id=self.partition_id)
+
     def join(
         self,
         right: vPartition,
@@ -284,6 +391,66 @@ class vPartition:
         output = vPartition(columns=result_columns, partition_id=self.partition_id)
         return output.eval_expression_list(output_schema)
 
+    def _to_file(
+        self,
+        file_format: str,
+        root_path: str,
+        partition_cols: Optional[ExpressionList] = None,
+        compression: Optional[str] = None,
+    ) -> List[str]:
+        keys = [col_id for col_id in self.columns.keys()]
+        names = [self.columns[k].column_name for k in keys]
+        data = [self.columns[k].block.to_arrow() for k in keys]
+        arrow_table = pa.table(data, names=names)
+        partition_col_names = []
+        if partition_cols is not None:
+            for col in partition_cols:
+                assert isinstance(
+                    col, ColumnExpression
+                ), "we can only support ColumnExpressions for partitioning parquet"
+                col_id = col.get_id()
+                assert col_id is not None
+                assert col_id in keys
+                partition_col_names.append(self.columns[col_id].column_name)
+
+        visited_paths = []
+
+        def file_visitor(written_file):
+            visited_paths.append(written_file.path)
+
+        format: pada.FileFormat
+        opts = None
+
+        if file_format == "parquet":
+            format = pada.ParquetFileFormat()
+            opts = format.make_write_options(compression=compression)
+        elif file_format == "csv":
+            format = pada.CsvFileFormat()
+            assert compression is None
+
+        pada.write_dataset(
+            arrow_table,
+            base_dir=root_path,
+            basename_template=str(uuid4()) + "-{i}." + format.default_extname,
+            format=format,
+            partitioning=partition_col_names,
+            file_options=opts,
+            file_visitor=file_visitor,
+            use_threads=False,
+            existing_data_behavior="overwrite_or_ignore",
+        )
+        return visited_paths
+
+    def to_parquet(
+        self, root_path: str, partition_cols: Optional[ExpressionList] = None, compression: Optional[str] = None
+    ) -> List[str]:
+        return self._to_file("parquet", root_path=root_path, partition_cols=partition_cols, compression=compression)
+
+    def to_csv(
+        self, root_path: str, partition_cols: Optional[ExpressionList] = None, compression: Optional[str] = None
+    ) -> List[str]:
+        return self._to_file("csv", root_path=root_path, partition_cols=partition_cols, compression=compression)
+
     @classmethod
     def merge_partitions(cls, to_merge: List[vPartition], verify_partition_id: bool = True) -> vPartition:
         assert len(to_merge) > 0
@@ -310,10 +477,20 @@ PartitionT = TypeVar("PartitionT")
 
 
 class PartitionSet(Generic[PartitionT]):
-    @abstractmethod
-    def to_pandas(self, schema: Optional[ExpressionList] = None) -> pd.DataFrame:
+    def _get_all_vpartitions(self) -> List[vPartition]:
         raise NotImplementedError()
 
+    def to_pydict(self) -> Dict[str, Sequence]:
+        """Retrieves all the data in a PartitionSet as a Python dictionary. Values are the raw data from each Block."""
+        all_partitions = self._get_all_vpartitions()
+        merged_partition = vPartition.merge_partitions(all_partitions, verify_partition_id=False)
+        return merged_partition.to_pydict()
+
+    def to_pandas(self, schema: Optional[ExpressionList] = None) -> "pd.DataFrame":
+        all_partitions = self._get_all_vpartitions()
+        part_dfs = [part.to_pandas(schema=schema) for part in all_partitions]
+        return pd.concat([pdf for pdf in part_dfs if not pdf.empty], ignore_index=True)
+
     @abstractmethod
     def get_partition(self, idx: PartID) -> PartitionT:
         raise NotImplementedError()
