@@ -1,13 +1,18 @@
 import contextlib
 import os
 from contextlib import contextmanager
+from typing import TYPE_CHECKING
 
 from loguru import logger
-from viztracer import VizTracer, get_tracer
+
+if TYPE_CHECKING:
+    from viztracer import VizTracer
 
 
-def profiler(filename: str) -> VizTracer:
+def profiler(filename: str) -> "VizTracer":
     if int(os.environ.get("DAFT_PROFILING", 0)) == 1:
+        from viztracer import VizTracer
+
         return VizTracer(output_file=filename, tracer_entries=10_000_000)
     else:
         return contextlib.nullcontext()
@@ -18,6 +23,8 @@ import time
 
 @contextmanager
 def timingcontext(name: str):
+    from viztracer import get_tracer
+
     tracer = get_tracer()
 
     try:
