@@ -1,10 +1,10 @@
 from __future__ import annotations
 
 from dataclasses import dataclass
-from typing import Any, Callable, ClassVar, Dict, List, Optional, Type
+from typing import Any, Callable, ClassVar, Dict, List, Optional, Type, cast
 
-import pandas as pd
 import ray
+from loguru import logger
 
 from daft.execution.execution_plan import ExecutionPlan
 from daft.execution.logical_op_runners import (
@@ -20,7 +20,6 @@ from daft.logical.optimizer import (
     PushDownLimit,
     PushDownPredicates,
 )
-from daft.logical.schema import ExpressionList
 from daft.resource_request import ResourceRequest
 from daft.runners.partitioning import PartID, PartitionManager, PartitionSet, vPartition
 from daft.runners.profiler import profiler
@@ -39,13 +38,11 @@ from daft.runners.shuffle_ops import (
 class RayPartitionSet(PartitionSet[ray.ObjectRef]):
     _partitions: Dict[PartID, ray.ObjectRef]
 
-    def to_pandas(self, schema: Optional[ExpressionList] = None) -> "pd.DataFrame":
+    def _get_all_vpartitions(self) -> List[vPartition]:
         partition_ids = sorted(list(self._partitions.keys()))
         assert partition_ids[0] == 0
         assert partition_ids[-1] + 1 == len(partition_ids)
-        all_partitions = ray.get([self._partitions[pid] for pid in partition_ids])
-        part_dfs = [part.to_pandas(schema=schema) for part in all_partitions]
-        return pd.concat([pdf for pdf in part_dfs if not pdf.empty], ignore_index=True)
+        return cast(List[vPartition], ray.get([self._partitions[pid] for pid in partition_ids]))
 
     def get_partition(self, idx: PartID) -> ray.ObjectRef:
         return self._partitions[idx]
@@ -206,7 +203,11 @@ class RayLogicalGlobalOpRunner(LogicalGlobalOpRunner):
 
 
 class RayRunner(Runner):
-    def __init__(self) -> None:
+    def __init__(self, address: Optional[str]) -> None:
+        if ray.is_initialized():
+            logger.warning(f"Ray has already been initialized, Daft will reuse the existing Ray context")
+        else:
+            ray.init(address=address)
         self._part_manager = PartitionManager(lambda: RayPartitionSet({}))
         self._part_op_runner = RayLogicalPartitionOpRunner()
         self._global_op_runner = RayLogicalGlobalOpRunner()
